//	basicsine~ -- make a basic sine wave (building towards a SPEAR external in class
//			Brad Garton, 12/2005


#include "ext.h"
#include "z_dsp.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define MAX_OUTPUTS 2	// also arbitrary

void *basicsine_class;

// declare the sine wave table out here so that all instances of the basicsine~ object
// can reference it; that way we don't create it anew for each object.
#define MAXSINETABLE 44100 // make it easy for us -- 1 cycle/second; si then equals freq
float sinetable[MAXSINETABLE];

// in this structure are variables specific to each instance of basicsine~
typedef struct _basicsine
{
	// header
	t_pxobject x_obj;

	long num_outputs; 	// number of outputs (mono or stereo)
	float si; // store the sampling increment used for the oscillator frequency
	float phase; // store where we currently are located in the sinetable[]
 
 } t_basicsine;


/****PROTOTYPES****/

//	max/msp setup funcs; this probably won't change, unless you decide to change the number of
//	args that the user can input, in which case basicsine_new will have to change
void *basicsine_new(long num_outputs);	// this allows us to have one optional params for the object
void basicsine_dsp(t_basicsine *x, t_signal **sp, short *count); 
t_int *basicsine_perform(t_int *w);
void basicsine_assist(t_basicsine *x, void *b, long m, long a, char *s);

// the following are functions we define below for handling floats/ints/bangs
// at the inputs.  We can delete these if we don't use them
void basicsine_float(t_basicsine *x, double f);
void basicsine_int(t_basicsine *x, int i);
void basicsine_bang(t_basicsine *x);


/****FUNCTIONS****/

//primary MSP funcs -- this is called the first time you instantiate an object of this type
void main(void)
{
	int i;
	float radians;
	
	// the A_DEFLONG argument sets mono or stereo output (default will be mono)
	setup((struct messlist **)&basicsine_class, (method)basicsine_new, (method)dsp_free, (short)sizeof(t_basicsine), 0L, A_DEFLONG, 0);

	// max/msp standard messages; don't change these  
	addmess((method)basicsine_dsp, "dsp", A_CANT, 0);
	addmess((method)basicsine_assist,"assist", A_CANT,0);

	// basicsine_float will be called for any floats coming in any inlet
	addfloat((method)basicsine_float);

	// basicsine_int will be called for any ints coming in any inlet
	addint((method)basicsine_int);

	// basicsine_bang will be called for any bangs coming in any inlet
	addbang((method)basicsine_bang);

	// msp initialization (gotta have this)
	dsp_initclass();

	// set up the sine wavetable
	for (i = 0; i < MAXSINETABLE; i++) {
		radians = (float)i/(float)MAXSINETABLE * TWOPI;
		sinetable[i] = sin(radians);
	}

	post("basicsine~ external loaded!");
}


// this gets called when the object is created and every time the user types in new args
void *basicsine_new(long num_outputs)
{
	int i;

	// leave this; creates our object
	t_basicsine *x = (t_basicsine *)newobject(basicsine_class);

	// zero out the struct, to be careful (takk to jkclayton)
	if (x) { 
		for (i = sizeof(t_pxobject); i < sizeof(t_basicsine); i++)  
			((char *)x)[i] = 0;
	}

	// this is where we access the values (if any) of the optional params for the object	
	// constrain number of outputs
	if (num_outputs < 1) num_outputs = 1;
	if (num_outputs > MAX_OUTPUTS) num_outputs = 1;

	// and we set the internal t_basicsine structure variables
	x->num_outputs = num_outputs;

	// now setup audio inputs and outputs
	// inputs
	dsp_setup((t_pxobject *)x, 0);
   
	// signal outputs
	for (i = 0; i < x->num_outputs; i++) {
		outlet_new((t_object *)x, "signal");
	}

	// initialize si and phase for this instance
	x->si = 0.0;
	x->phase = 0.0;

	//occasionally this line is necessary if you are doing weird asynchronous things with the in/out vectors
	//x->x_obj.z_misc = Z_NO_INPLACE;

	return (x);
}


// this gets called everytime audio is started; even when audio is running, if the user changes anything
// (like deletes a patch cord), audio will be turned off and then on again, calling this func.
// this adds the "perform" method to the DSP chain, and also tells us
// where the audio vectors are and how big they are
void basicsine_dsp(t_basicsine *x, t_signal **sp, short *count)
{
	void *dsp_add_args[MAX_OUTPUTS + 2];
	int i;

	// construct the array of vectors and stuff for max/msp to knoww
	dsp_add_args[0] = x; // the object itself
	for (i = 0; i < (x->num_outputs); i++) { // pointers to the input and output vectors
		dsp_add_args[i+1] = sp[i]->s_vec;
	}
	dsp_add_args[x->num_outputs + 1] = (void *)sp[0]->s_n; // pointer to the vector size
	dsp_addv(basicsine_perform, (x->num_outputs + 2), dsp_add_args); // add them to the signal chain	
}


// this is where the action is
// we get vectors of samples (n samples per vector), process them and send them out
// the "w" array contains a lot of stuff (note no input pointers in this one!)
//	w[0] -- pointer to the basicsine_perform function being called
//	w[1] -- pointer to the t_basicsine object structure data
//	w[2] ... w[numoutputs+1] -- pointers to arrays to be filled with samples (output arrays)
//	w[numoutputs+2] -- number of samples/vector (i.e. the number of samples in each input or output array)
t_int *basicsine_perform(t_int *w)
{
	t_basicsine *x = (t_basicsine *)(w[1]);

	float *out[MAX_OUTPUTS];	// pointers to the output vectors
	// note the the above are pointers -- each output is represented as an array
	// of samples (i.e. non-interleaved).  These are set to point to the w[] arrays.

	long n = w[x->num_outputs + 2];	// number of samples per vector

	// local vars
	int i;

	// check to see if we should skip this routine if the patcher is "muted"
	// all audio objects should have this, even though the 'goto' is supremely elegant.
	if (x->x_obj.z_disabled) goto out;

	// assign the output vector pointers
	for (i = 0 ;i < x->num_outputs; i++) {
		out[i] = (float *)(w[i+2]);
	}
	// after this loop above, we can write samples using *out[oulet]+sampval to
	// get and set sample values for each outlet vector.


/* ------- this is the sample computing loop for each vector ------- */
	// we're going to count down by the total # of samples, and process each of the current output
	// vectors.  Alternatively we could have counted through each vector separately, but we didn't
	while (n--) { // n is the number of samples for the vector
		*out[0]++ = sinetable[(int)x->phase]; // we are ignoring fractional si (i.e. non-interpolating oscillator)
		if (x->num_outputs == 2) *out[1]++ = sinetable[(int)x->phase]; //two channel output, dumb for now

		// increment our phase pointer, and wrap around to the beginning of the sinetable if necessary
		x->phase += x->si;
		if (x->phase > (float)MAXSINETABLE) x->phase = x->phase - (float)MAXSINETABLE;
	}
/* ------- end of the sample computing loop ------- */


	// return a pointer to the next object in the signal chain.
out:
	return w + x->num_outputs + 3;
}


// tells the user about the inputs/outputs when mousing over them
// "m" is inlet (1) or outlet (2); "a" is which one the mouse is over
void basicsine_assist(t_basicsine *x, void *b, long m, long a, char *s)
{
	int i;
	
	//could use switch/case inside for loops, to give more informative assist info....
	if (m == 1) sprintf(s, "frequency");  // this is the message that gets shown in the window frame
	if (m == 2) {
		for (i = 0; i <= x->num_outputs; i++) // "<=" because we have one additional outlet for bangs/vals
			if (a == i) sprintf(s, "signal %d", i);  // ditto about message that gets shown
	}
}


// we'll use float and int input to set the frequency
// this gets called whenever a float is received at *any* input
void basicsine_float(t_basicsine *x, double f)
{
	x->si = f; // this is because we are using a sinetable of 44100 length
}


// this gets called whenever an int is received at *any* input
void basicsine_int(t_basicsine *x, int k)
{
	x->si = (float)k;  // see above
}


// this gets called whenever a bang is received at *any* input
void basicsine_bang(t_basicsine *x)
{
	post("this is the basicsine~: object");
}
